from .pattern import Pattern


class Factor:
    """Implementation of a theme factor."""

    def __init__(
        self,
        id: str,
        rate: float,
        patterns: list[Pattern],
    ):
        """Initialisation of the class.

        Parameters
        ----------
        id: int
            name of the factor
        rate: float
            how often an answer is given
        patterns: list of Patterns
            list of time repeating patterns
        """

        self.id: str = id
        self.rate: float = rate
        self.patterns: list[Pattern] = patterns

    def __str__(self):
        return f"{self.__class__.__name__} '{self.name}'"

    def __repr__(self):
        return f"<Factor {id}"

    @classmethod
    def load(cls, item: dict):
        """Create class instance from dictionnary.

        Parameters
        ----------
        item: dict
            input to create the instance from.
        """

        factor = Factor(
            item.get("id", "0.0"),
            item.get("rate", 1),
            [Pattern.load(pattern) for pattern in item.get("patterns", [])],
        )

        return factor



from .question import Question


class Theme:
    """Implementation of a theme containing questions."""

    def __init__(self, name: str):
        """Initialisation of the class.

        Parameters
        ----------
        name: str
            name of the theme
        """

        self.name: str = name
        self.questions: Question = []

    def __str__(self):
        return f"<Theme> '{self.name}': {len(self.questions)} questions"

    def __repr__(self):
        return self.__str__()



class Question:
    """Implementation of a question."""

    def __init__(self, text: str):
        """Initialisation of the class.

        Parameters
        ----------
        text: str
            content of the question
        """
        self.text: str = text

    def __str__(self):
        return f"<{self.__class__.__name__}> {self.text}"

    def __repr__(self):
        return self.text



import random

from .persona import Persona


class Group:
    """Implementation of a group composed of people."""

    def __init__(self, name: str, personas: list[Persona], size: int):
        """Initialisation of the class.

        Parameters
        ----------
        name: str
            name of the group
        personas: list of Persona
            people's kinds to generate the group from
        size: int
            size of the group
        """
        self.name: str = name
        self.personas: list[Persona] = personas
        self.size: int = size
        self.people: list[Persona] = [random.choice(self.personas) for _ in range(self.size)]
        
    @property
    def user_names(self):
        return [p.name for p in self.people]
        
    def __str__(self):
        names = [persona.name for persona in self.personas]
        return f"'{self.name.upper()}' ({self.size} people) from {names}"

    def __repr__(self):
        return f"<{self.__class__.__name__}> '{self.name}'"

    @classmethod
    def load(cls, item: dict):
        """Create class instance from dictionnary.

        Parameters
        ----------
        item: dict
            input to create the instance from.
        """

        group_name: str = item.get("meta", {}).get("name", "default")
        group_size: int = item.get("meta", {}).get("size", 0)
        persona_list: list[dict] = item.get("personas", {})

        personas: list[Persona] = [Persona.load(p) for p in persona_list]
        
        return cls(group_name, personas, group_size)

            


from .datagenerator import DataGenerator
from .group import Group
from .pattern import Pattern
from .persona import Persona
from .factor import Factor
from .question import Question
from .themerating import ThemeRating
from .logger import Logger

__all__ = [
    "Persona",
    "Question",
    "ThemeRating",
    "Pattern",
    "Factor",
    "Group",
    "DataGenerator",
    "Logger",
]



from datetime import datetime
from tkinter import N

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import collections

from .logger import Logger
from .persona import Persona
from .themerating import ThemeRating


class DataGenerator:
    """Implementation of a generator based on a persona rating."""

    colors = ["g", "b", "r", "k", "orange"]

    def __init__(self, startdate: datetime, rating: ThemeRating, samples: int):
        """Initialisation of the class.

        Parameters
        ----------
        startdate : datetime
            The date of the first sample.
        rating: ThemeRating
            How a person would react to a theme.
        samples: int
            Length of the data to generate.
        """

        self.startdate = startdate
        self.rating = rating
        self.samples = samples
        self.data = np.ones([2, samples]) * (-1)

    @property
    def mean(self) -> float:
        """Return the mean of the generated data."""

        return np.nanmean(self.data[1, :])

    @property
    def rate(self) -> float:
        """Return the fill rate of the generated data."""

        return np.sum(~np.isnan(self.data[1, :])) / self.data.shape[1]

    def fill(self):
        """Fill the data in the correct order."""

        self._respect_pattern()
        self._respect_mean()
        self._respect_factor_rate()
        self._respect_fill_rate()

    def show(self, person: Persona):
        """Show the generated data."""

        if not Logger.verbose:
            return

        Logger.write(
            f"[{'GENERATOR':10s}]     Generated average: {self.mean:.2f} (req {self.rating.average:.2f})"
        )

        Logger.write(
            f"[{'GENERATOR':10s}]     Generated fill-rate: {self.rate:.2f} (req {self.rating.rate:.2f})"
        )

        n_factors = len(self.rating.factors)

        _, ax = plt.subplots(figsize=(12, 5))

        for factor in range(n_factors):

            factor_indexes = np.where(self.data[0, :] == factor)[0]
            factor_data = self.data[1, factor_indexes]

            lines = [
                [(idx, 0), (idx, value)]
                for idx, value in zip(factor_indexes, factor_data)
            ]

            linecoll = collections.LineCollection(
                lines, linewidths=1, color=DataGenerator._get_color(factor)
            )
            ax.add_collection(linecoll)

            plt.scatter(
                factor_indexes,
                factor_data,
                s=6,
                color=DataGenerator._get_color(factor),
                label=f"factor {factor} ({self.rating.factors[factor].rate:.0%})",
            )

        plt.axhline(y=self.mean, color="r", linestyle="--", label="data mean")
        plt.axhline(
            y=self.rating.average, color="g", linestyle="--", label="required mean"
        )
        plt.yticks(self.rating.possible_values + [0])
        plt.xticks(np.arange(0, np.ceil(self.samples / 10) * 10 + 1, 10))
        plt.grid()

        plt.title(f"Data generated for {person.name} (theme {self.rating.id})")
        plt.legend()

        plt.show()

    @classmethod
    def _get_color(cls, index: int):
        """Get a color in a cyclic maner."""

        return DataGenerator.colors[index % len(DataGenerator.colors)]

    def _respect_pattern(self):
        """Fill the data according to the patterns specified by the rating."""

        for factor_idx, factor in enumerate(self.rating.factors):
            for pattern in factor.patterns:
                indexes = np.arange(pattern.start, self.samples, pattern.step)
                self.data[0, indexes] = factor_idx
                self.data[1, indexes] = pattern.value

    def _respect_mean(self):
        """Fill the remaining data so that the average value equals the one in the rating."""

        patterns = [p for factor in self.rating.factors for p in factor.patterns]
        mean = (self.rating.average - np.sum([p.value / p.step for p in patterns])) / (
            1 - np.sum([1 / p.step for p in patterns])
        )

        indexes = np.where(self.data[0, :] == -1)[0]

        self.data[1, indexes] = self._get_randoms(mean, indexes.shape[0])

    def _respect_factor_rate(self):
        """Fill the data according to the patterns specified by the rating."""

        splits_counts = [
            int(factor.rate * self.data.shape[1]) for factor in self.rating.factors
        ]

        for idx, split_count in enumerate(splits_counts):
            remaining_indexes = np.where(self.data[0, :] == -1)[0]

            n_to_pick = split_count - np.count_nonzero(self.data[0, :] == idx)
            indexes = np.random.choice(remaining_indexes, n_to_pick, replace=False)

            self.data[0, indexes] = idx

    def _respect_fill_rate(self):
        """Fill the data so that the filling rate is respected."""

        n_to_remove = round(self.samples * (1 - self.rating.rate))
        indexes = np.random.choice(np.arange(self.samples), n_to_remove, replace=False)

        self.data[0, indexes] = None
        self.data[1, indexes] = None

    def _get_randoms(self, mean: float, size: int = 0) -> np.ndarray:
        """Get random values with a given mean."""

        computed_mean: float = None

        while not computed_mean or not (mean + 0.01 >= computed_mean >= mean - 0.01):
            rand = np.random.choice(self.rating.possible_values, size)
            computed_mean = rand.mean()

        return rand

    def __str__(self):
        return



class Pattern:
    """Implementation of how a pattern will repeat in time."""

    def __init__(self, step: int, start: int, value: int):
        """Initialisation of the class.


        Parameters
        ----------
        step: int
            how often the patterns repeats
        start: int
            when the patterns starts
        value: int
            value to repeat
        """

        self.step: int = step
        self.start: int = start
        self.value: int = value

    def __repr__(self):
        return f"<{self.__class__.__name__}> step:{self.step}, start: {self.start}, value: {self.value}"

    def __str__(self):
        return f"step:{self.step}, start: {self.start}, value: {self.value}"

    @classmethod
    def load(cls, item: dict):
        """Create class instance from dictionnary.

        Parameters
        ----------
        item: dict
            input to create the instance from.
        """

        return Pattern(
            item.get("step", 0),
            item.get("start", 0),
            item.get("value", 0),
        )



from .theme import Theme
from .factor import Factor


class ThemeRating:
    """Implementation of how a persona would react to a given theme."""

    def __init__(
        self,
        id: str,
        rate: float,
        average: float,
        possible_values: list[int],
        factors: list[Factor],
    ):
        """Initialisation of the class.

        Parameters
        ----------
        id: str
            theme to react to
        rate: float
            how often an answer is given
        average: float
            mean value of the answers
        possible_values: list of int
            list of values that can be used
        factors: list of Factors
            list of factors
        """

        self.id: str = id
        self.rate: float = rate
        self.average: float = average
        self.possible_values: list[int] = possible_values
        self.factors: list[Factor] = factors

    def __str__(self):
        return f"{self.__class__.__name__} '{self.theme}'"

    def __repr__(self):
        return f"<Rating for '{self.theme}' theme"

    @classmethod
    def load(cls, item: dict):
        """Create class instance from dictionnary.

        Parameters
        ----------
        item: dict
            input to create the instance from.
        """

        rating = ThemeRating(
            item.get("id", "0"),
            item.get("rate", 1),
            item.get("average", 0),
            item.get("possible_values", []),
            [Factor.load(pattern) for pattern in item.get("factors", [])],
        )

        return rating



from .themerating import ThemeRating


class Persona:
    """Implementation of a fictional target"""

    def __init__(self, name: str, description: str, ratings: list[ThemeRating]):
        """Initialisation of the class.

        Parameters
        ----------
        name: str
            name of the persona
        description: str
            what are the persona's main characteristics
        ratings: list of ThemeRating
            to what themes and hiw is the persona answering
        """

        self.name: str = name
        self.description: str = description
        self.ratings: list[ThemeRating] = ratings

    def __str__(self):
        text = f"{self.name}:  {self.description} (rating {len(self.ratings)} themes)"
        return text

    def __repr__(self):
        return f"<{self.__class__.__name__}> '{self.name}': {self.ratings}"

    @classmethod
    def load(cls, item: dict):
        """Create class instance from dictionnary.

        Parameters
        ----------
        item: dict
            input to create the instance from.
        """

        name: str = item.get("name", "default")
        description: str = item.get("description", "-")
        ratings_dict = item.get("ratings", [])

        ratings = [ThemeRating.load(rating) for rating in ratings_dict]

        return cls(name, description, ratings)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="spoton-generator",
    version="0.1",
    scripts=["bin/spoton-generator"],
    author="Jean Demeusy",
    author_email="dev.jdu@gmail.com",
    description="A tool to generate data for Spot-On",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.forge.hefr.ch/pa-spot-on-demeusy/spot-on-generator",
    packages=["spoton-generator"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["spoton-soozaccess"],
)



import json
from datetime import datetime

import click
from soozaccess import SoozAccess, daterange

from .datagenerator import DataGenerator
from .group import Group
from .logger import Logger


@click.command()
@click.option(
    "-f", "--filename", type=str, help="path to json data file.", required=True
)
@click.option(
    "-s", "--samples", type=int, help="number of samples to generate", required=True
)
@click.option(
    "-h", "--host", type=str, help="url to host", default="http://localhost:3000"
)
@click.option(
    "-v", "--verbose", count=True, help="show results and interface", default=False
)
@click.option(
    "-d",
    "--startdate",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="start date to generate data for",
)
def generator(
    filename: str,
    host: str,
    samples: int,
    verbose: bool,
    startdate: datetime,
):
    """Main entry for SooZ data generator."""

    # create log instance
    Logger.verbose = verbose

    # import group personas from json file
    with open(filename, "r") as f:
        group = Group.load(json.load(f))

    # prepare database
    # SoozAccess.verbose = verbose
    # SoozAccess.host = host
    # SoozAccess.prepare_users(group.name, group.user_names)

    # display
    Logger.write(
        f"[{'GENERATOR':10s}] Imported group '{group.name}' with {len(group.people)} people from {len(group.personas)} personas."
    )

    # data generation
    for p in group.personas:
        for rating in p.ratings:
            Logger.write(f"[{'GENERATOR':10s}] {p.name}. rating id: {rating.id}")

            datagen = DataGenerator(startdate, rating, samples)
            datagen.fill()
            datagen.show(p)

            # with SoozAccess(p.name, p.name) as user:
            #     for day in daterange(startdate, samples):
            #         user.put_response(2, 2, 3, day.strftime("%Y-%m-%d"))


if __name__ == "__main__":
    generator()
